/*
 * Raid 2: The Return of Saddam
 * Copyright (c) 2008 Steven Wallace / Chedda Cheeze
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

// Music interface
// Note: there is no interface to remove incomplete downloads yet!
#include "Music.h"
#include <stdio.h>
#include <curl/curl.h>
#include <allegro.h>
#include <alogg/alogg.h>
#include <pthread.h>


char *Music::StatusString() {
#ifdef NO_MUSIC
	return "MUSIC DISABLED BY COMPILER FLAG. REMOVE `-DNO_MUSIC' FROM THE MAKEFILE TO ENABLE IT.";
#else
	static char status[1024];
	if(sample && fetch_idx >= 0) 
		snprintf(status, 1024, "Now Playing \"%s\"; Downloading \"%s\" (%0.1f%%)", song[playing_idx]->title, song[fetch_idx]->title, fetch_progress);
	else if(sample)
		snprintf(status, 1024, "Now Playing \"%s\"", song[playing_idx]->title);
	else if(fetch_idx >= 0)
		snprintf(status, 1024, "Downloading \"%s\" (%0.1f%%)", song[fetch_idx]->title, fetch_progress);
	else if(playing_idx >= 0)
		snprintf(status, 1024, "Loading \"%s\"", song[playing_idx]->title);
	else
		snprintf(status, 1024, "Music player is idle.");
	
	return status;
#endif
}


int FetchCallback(void *data, double dltotal, double dlnow, double ultotal, double ulnow) {
	Music *m = (Music*)data;
	m->fetch_progress = dlnow/dltotal*100;
	return 0;
}


// This function must be run once in a separate thread
// It will fetch all music and then kill itself
void *FetchThread(void *ptr) {
	Music *m = (Music*)ptr;
	
	int i;
	for(i=0; i<m->song.size(); i++) {
		// Check if we need this song downloaded
		Song *s = m->song[i];
		if(s->available) continue;
		m->fetch_idx = i;
		
		// Initialize cURL
		CURL *handle;
		char errstr[1024];
		curl_global_init(CURL_GLOBAL_ALL);
		
		// Open the file for output
		FILE *f = fopen(s->filename, "wb");
		if(!f) continue;
		
		// Set cURL options
		handle = curl_easy_init();
		curl_easy_setopt(handle, CURLOPT_WRITEDATA, f);
		curl_easy_setopt(handle, CURLOPT_NOPROGRESS, 0);
		curl_easy_setopt(handle, CURLOPT_FOLLOWLOCATION, 1);
		curl_easy_setopt(handle, CURLOPT_PROGRESSFUNCTION, &FetchCallback);
		curl_easy_setopt(handle, CURLOPT_PROGRESSDATA, ptr);
		curl_easy_setopt(handle, CURLOPT_ERRORBUFFER, errstr);
		
		// Download the file
		curl_easy_setopt(handle, CURLOPT_URL, s->url);
		
		if(curl_easy_perform(handle) != 0) continue;
		curl_easy_cleanup(handle);
		fclose(f);
		s->available = true;
	}
	
	m->fetch_idx = -1;
	
	return 0;
}


Music::Music(const char *filename) {
#ifndef NO_MUSIC
	// Set some initial variables
	fetch_idx = -1;
	sample = NULL;
	voice = -1;
	playing_idx = -1;
	
	// Open specified playlist file
	FILE *f = fopen(filename, "rt");
	if(!f) return;
	
	
	bool geturl=true;
	char url[1024];
	while(!feof(f)) {
		char line[1024];
		fgets(line, 1024, f);
		if(line[0] == '#' || line[0] == ' ' || line[0] == '\n' || line[0] == '\t') continue;
		
		// Remove the '\n' at the end of each line
		line[strlen(line)-1]='\0';
		
		if(geturl) snprintf(url, 1024, "%s", line);
		else {
			Song *s = new Song(url, line);
			song.push_back(s);
			if(s->available) available_song.push_back(s);
		}
		
		// Alternate if we're getting a url or a title
		geturl = !geturl;
	}
	
	pthread_create(&fetch_thread, NULL, &FetchThread, (void*)this);
	
	fclose(f);
#endif
}


void Music::PlayFirstSong() {
	PlaySong(0);
}

void Music::PlayRandomSong() {
	if(available_song.size() > 0) {
		int idx = rand()%(available_song.size()-1);
		if(idx == playing_idx) idx++;
		if(idx >= available_song.size()) idx = 0;
		PlaySong(idx);
	}
}

void Music::PlayLastSong() {
	PlaySong(available_song.size()-1);
}


void *PlaySongThread(void *parent) {
	Music *m = (Music*)parent;
	Song *s = m->song[m->playing_idx];

	// If something is already playing, stop it
	if(m->sample) {
		voice_stop(m->voice);
		deallocate_voice(m->voice);
		destroy_sample(m->sample);
	}
	
	// Load and play the sample
	fprintf(stderr, "Loading %s...\n", s->title);
	m->sample = alogg_load_ogg(s->filename);
	if(m->sample) {
		m->voice = allocate_voice(m->sample);
		voice_start(m->voice);
		voice_set_priority(m->voice, 255);	// Make sure this voice is never killed by anyone but us
	}
	
	return 0;
}

void Music::PlaySong(int idx) {
#ifndef NO_MUSIC
	// Make sure the index is valid
	if(idx >= song.size()) return;
	
	// Make sure the song is available
	Song *s = song[idx];
	if(!s->available) return;
	
	playing_idx = idx;
	timer = 0.0;
	
	// Spawn thread to load the song
	//load_thread = 0;
	pthread_create(&load_thread, NULL, &PlaySongThread, (void*)this);	
#endif
}


void Music::PollMusic(float dt) {
#ifndef NO_MUSIC
	// Give it 30 seconds before trying to load something else
	timer += dt;
	if(sample && voice_check(voice) == sample) return;
	else if(timer > 30.0) PlayRandomSong();
#endif
}



Song::Song(const char *urlin, const char *titlein) {
	// Copy the url and title
	snprintf(url, 1024, "%s", urlin);
	snprintf(title, 1024, "%s", titlein);
	

	// Determine the filename from the URL
	filename = rindex(url, '/');
	if(filename == NULL); // Uh-oh, what do I do here?
	filename++;
	
	
	// Check if the file is there
	if(file_size_ex(filename) > 0) available = true;
	else available = false;
}
